/*
** Copyright (c) 2003, Matt Messier and John Viega
** All rights reserved.
** 
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
** 
** Redistributions of source code must retain the above copyright notice,
** this list of conditions and the following disclaimer.
** 
** Redistributions in binary form must reproduce the above copyright
** notice, this list of conditions and the following disclaimer in the
** documentation and/or other materials provided with the distribution.
** 
** Neither the name of the primary nor the names of the contributors may
** be used to endorse or promote products derived from this software
** without specific prior written permission.
** 
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef WIN32
#include "config.h"
#include "safestr.h"

#if defined(STDC_HEADERS)
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#else
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0400
#endif
#include <windows.h>
#include <wincrypt.h>
#include <limits.h>
#include "safestr.h.in"
#endif

#define XXL_ENFORCE_PREFIX
#include <xxl.h>

#ifdef WIN32
typedef unsigned __int64   u_int64_t;
#else
#  ifndef HAVE_U_INT64_T
#    ifdef HAVE_LONG_LONG
typedef unsigned long long u_int64_t;
#    else
#      error Need a definition for u_int64_t!
#    endif
#  endif
#endif

#define MOD_REF_COUNT(x, y)                                                 \
            (x)->hdr.flags = (((REF_COUNT(x) + (y)) << 16) & 0xffff0000) |  \
                             ((x)->hdr.flags & 0x0000ffff)

#define REF_COUNT(x)      (((x)->hdr.flags >> 16) & 0xffff)
#define INC_REF_COUNT(x)  MOD_REF_COUNT(x, 1)
#define DEC_REF_COUNT(x)  MOD_REF_COUNT(x, -1)

typedef struct small_isafestr_t
{
    /* Number of characters you have room for in the string.  That is, this
     * number is the total allocated size minus the length of the header and
     * the byte for the NUL terminating character.
     */
    u_int32_t           size;     
    u_int32_t           length;   
    u_int32_t           flags;
    u_int32_t           cookie;
} small_isafestr_t;

typedef struct isafestr_t
{
    small_isafestr_t    hdr;
    char                str[1];
} *isafestr_t;

#define SAFESTR_ROUND(x)    ((((x) + 127) / 128) * 128)
#define SAFESTR_MIN(a, b)   ((a) > (b) ? (b) : (a))

#define SAFESTR_FMTARGS     256

#ifdef WIN32
#define strcmp(s1, s2)          (CompareStringA(GetUserDefaultLCID(), 0, (s1), -1, (s2), -1) - 2)
#define strncmp(s1, s2, n)      (CompareStringA(GetUserDefaultLCID(), 0, (s1), (n), (s2), (n)) - 2)
#define strlen(s1)              lstrlenA((s1))
#define strcasecmp(s1, s2)      (CompareStringA(GetUserDefaultLCID(), NORM_IGNORECASE, (s1), -1, (s2), -1) - 2)
#define strncasecmp(s1, s2, n)  (CompareStringA(GetUserDefaultLCID(), NORM_IGNORECASE, (s1), (n), (s2), (n)) - 2)
#endif

static u_int32_t         safestr_cookie     = 0;
static safestr_malloc_t  safestr_malloc_fn  = malloc;
static safestr_realloc_t safestr_realloc_fn = realloc;
static safestr_free_t    safestr_free_fn    = free;

static void       *safestr_malloc(size_t, xxl_assettype_t type);
static void       free_isafestr_asset(isafestr_t, void *);
static void       free_void_asset(void *, void *);
static isafestr_t get_isafestr_ro(safestr_t);
static isafestr_t get_isafestr_rw(safestr_t, int);
static isafestr_t resize_safestr(isafestr_t, u_int32_t);
static void       bake_cookie(void);
static u_int32_t  get_cookie(void);
static int        basecmp(safestr_t, safestr_t, u_int32_t, u_int32_t);
static safestr_t  *charlist(isafestr_t str);
static int        check_format_string(isafestr_t, va_list);

static void *
safestr_malloc(size_t size, xxl_assettype_t type)
{
    void    *ptr;

    if (!(ptr = safestr_malloc_fn(size)))
        XXL_THROW_ERROR(SAFESTR_ERROR_OUT_OF_MEMORY, NULL);
    XXL_ASSET_SAVE(ptr, free_void_asset, NULL, type);

    return ptr;
}

static void
free_isafestr_asset(isafestr_t str, void *arg)
{
    DEC_REF_COUNT(str);
    if (REF_COUNT(str) <= 0 ||
        (REF_COUNT(str) == 1 && (str->hdr.flags & SAFESTR_TEMPORARY)))
    {
        if (str->hdr.cookie != safestr_cookie)
            XXL_THROW_ERROR(SAFESTR_ERROR_BAD_ADDRESS, NULL);
#ifdef DEBUG
        if (getenv("SAFESTR_TRACE"))
        {
            fprintf(stderr, "Freeing string 0x%08X (\"%s\")\n",
                    (unsigned int)str->str, str->str);
        }
#endif
        safestr_memzero(str->str, str->hdr.size);
        str->hdr.size = str->hdr.length = str->hdr.flags = str->hdr.cookie = 0;
        safestr_free_fn(str);
    }
}

static void
free_void_asset(void *ptr, void *arg)
{
    safestr_free_fn(ptr);
}

static isafestr_t
get_isafestr_ro(safestr_t s)
{
    isafestr_t  str;

    if (!s)
        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_PARAMETER, NULL);
    str = (isafestr_t)((char *)s - sizeof(small_isafestr_t));
    if (str->hdr.cookie != safestr_cookie)
        XXL_THROW_ERROR(SAFESTR_ERROR_BAD_ADDRESS, NULL);
    INC_REF_COUNT(str);
    XXL_ASSET_SAVE(str, free_isafestr_asset, NULL, XXL_ASSET_TEMPORARY);
    return str;
}

static isafestr_t
get_isafestr_rw(safestr_t s, int trust)
{
    isafestr_t str;

    str = get_isafestr_ro(s);
    if (str->hdr.flags & SAFESTR_IMMUTABLE)
        XXL_THROW_ERROR(SAFESTR_ERROR_IMMUTABLE_STRING, s);
    switch (trust)
    {
        case -1:  /* Flag is unchanged. */
            break;
        case 0:
            str->hdr.flags &= ~SAFESTR_TRUSTED;
            break;
        default:
            str->hdr.flags |= SAFESTR_TRUSTED;
            break;
    }
    return str;
}

static isafestr_t
resize_safestr(isafestr_t str, u_int32_t length)
{
    u_int32_t   size;
    isafestr_t  new_str;

    if (length <= str->hdr.size)
    {
        safestr_memzero(str->str + length, str->hdr.size - length);
        return str;
    }
    if (length + 1 > (1 << (sizeof(u_int32_t) * 4)) - sizeof(small_isafestr_t))
        XXL_THROW_ERROR(SAFESTR_ERROR_STRING_TOO_LONG, NULL);
    size = SAFESTR_ROUND(sizeof(small_isafestr_t) + length + 1);

    new_str = (isafestr_t)safestr_malloc(size, XXL_ASSET_PERMANENT);
    new_str->hdr.size   = size - sizeof(small_isafestr_t) - 1;
    new_str->hdr.length = length;
    new_str->hdr.flags  = str->hdr.flags;
    new_str->hdr.cookie = str->hdr.cookie;
    memcpy(new_str->str, str->str, str->hdr.length);
    safestr_memzero(str->str, str->hdr.size);
    str->hdr.size = str->hdr.length = str->hdr.cookie = str->hdr.flags = 0;
    safestr_free_fn(str);
    XXL_ASSET_UPDATE(str, new_str);
    return new_str;
}

static void
bake_cookie(void)
{
#ifdef WIN32
    HCRYPTPROV  c;

    if (!CryptAcquireContext(&c, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) 
        XXL_THROW_ERROR(SAFESTR_ERROR_PRNG_FAILURE, NULL);
    if (!CryptGenRandom(c, sizeof(safestr_cookie), (BYTE *)&safestr_cookie))
    {
        CryptReleaseContext(c, 0);
        XXL_THROW_ERROR(SAFESTR_ERROR_PRNG_FAILURE, NULL);
    }
    CryptReleaseContext(c, 0);
#else
    int fd;

    if ((fd = open("/dev/urandom", O_RDONLY)) == -1)
        XXL_THROW_ERROR(SAFESTR_ERROR_PRNG_FAILURE, NULL);
    read(fd, &safestr_cookie, sizeof(safestr_cookie));
    close(fd);
#endif
}

static u_int32_t
get_cookie(void)
{
    if (!safestr_cookie)
        bake_cookie();
    return safestr_cookie;
}

static int
basecmp(safestr_t s1, safestr_t s2, u_int32_t flags, u_int32_t length)
{
    int         result;
    isafestr_t  str1, str2;

    XXL_ASSET_BLOCK_BEGIN
    {
        str1 = get_isafestr_ro(s1);
        str2 = get_isafestr_ro(s2);

        if (length > -1)
        {
            if (!(flags & SAFESTR_COMPARE_NOCASE))
                result = strncmp(str1->str, str2->str, length);
            else
                result = strncasecmp(str1->str, str2->str, length);
        }
        else
        {
            if (!(flags & SAFESTR_COMPARE_NOCASE))
                result = strcmp(str1->str, str2->str);
            else
                result = strcasecmp(str1->str, str2->str);
        }
    }
    XXL_ASSET_BLOCK_END;

    return result;
}

static safestr_t *
charlist(isafestr_t str)
{
    safestr_t   *ret;
    u_int32_t   i, size;
    isafestr_t  istr;

    /* Assets saved in this function will be saved within the context of the
     * caller: safestr_split()
     */

    size = sizeof(safestr_t) * (str->hdr.length + 1);
    ret = (safestr_t *)safestr_malloc(size, XXL_ASSET_PROMOTE);

    for (i = 0;  i < str->hdr.length;  i++)
    {
        ret[i] = safestr_alloc(1, (str->hdr.flags & SAFESTR_TRUSTED));
        istr   = (isafestr_t)((char *)ret[i] - sizeof(small_isafestr_t));
        XXL_ASSET_SAVE(istr, free_isafestr_asset, NULL, XXL_ASSET_PROMOTE);
        istr->hdr.length = 1;
        istr->str[0] = str->str[i];
        istr->str[1] = '\0';
    }
    ret[str->hdr.length] = NULL;
    return ret;
}

static int
check_format_string(isafestr_t fmt, va_list ap)
{
    int        float_expected, i, l, modifiers, popi, seen_dsgn, trusted;
    char       argument_sizes[SAFESTR_FMTARGS] = { 0 }, c, *s;
    safestr_t  popc;
    u_int64_t  argnum, maxarg = 0, nextarg = 0;
    isafestr_t str;

    s = fmt->str;
    trusted = (fmt->hdr.flags & SAFESTR_TRUSTED) == SAFESTR_TRUSTED;
    while ((c = *s++)) 
    {
        if (c != '%')
            continue;

        argnum = 0;
        float_expected = l = modifiers = seen_dsgn = 0;

modifier_start:
        switch ((c = *s++))
        {
            case '#':
            case '-':
            case '+':
            case '0':
            case ' ':
                goto modifier_start;
            case '.':
                while ((*s) >= '0' && (*s) <= '9')
                    s++;
                goto modifier_start;
            case 'h':
                if (float_expected == 1 || (modifiers && modifiers != sizeof(short)))
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                float_expected = -1;
                modifiers = sizeof(short);
                goto modifier_start;
            case 'l':
                if (float_expected == 1)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                float_expected = -1;
                if (!l)
                { 
                    l = 1;
                    if (modifiers && modifiers != sizeof(long))
                        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                    modifiers = sizeof(long);
                }
                else
                {
                    if (modifiers != sizeof(long) && modifiers != sizeof(u_int64_t))
                        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                    modifiers = sizeof(u_int64_t);
                }
                goto modifier_start;
            case 'q':
#ifdef HAVE_QUAD_WORDS
                if (float_expected == 1 || (modifiers && modifiers != sizeof(quad int)))
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                float_expected = -1;
                modifiers = sizeof(quad int);
#else
                XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
#endif
                goto modifier_start;
            case 'L':
#ifdef HAVE_LONG_DOUBLE
                if (float_expected == -1 || (modifiers && modifiers != sizeof(long double)))
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                float_expected = 1;
                modifiers =  sizeof(long double);
#else
                XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
#endif
                goto modifier_start;
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                argnum = c - '0';
                while (*s >= '0' && *s <= '9')
                {
                    argnum = (argnum * 10) + (*s++ - '0');
                    if (argnum > INT_MAX)
                        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                }
                if (*s == '$')
                {
                    s++;
                    if (seen_dsgn)
                        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                    seen_dsgn = 1;
                    if (argnum > SAFESTR_FMTARGS)
                        XXL_THROW_ERROR(SAFESTR_ERROR_TOO_MANY_FORMAT_ARGS, NULL);
                }
                goto modifier_start;
            case 'D':
            case 'O':
            case 'U':
                if (float_expected == 1)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                if (!modifiers)
                    modifiers = sizeof(long);
                goto process_arg;
            case 'c':
                if (float_expected == 1)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                if (modifiers && modifiers != sizeof(int))
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                modifiers = sizeof(int);
                goto process_arg;
            case 's':
                if (modifiers || float_expected)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                modifiers = -1;
                goto process_arg;
            case 'p':
                if (float_expected || (modifiers && modifiers != sizeof(void *)))
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                modifiers = sizeof(void *);
                goto process_arg;
            case 'n':
                XXL_THROW_ERROR(SAFESTR_ERROR_ILLEGAL_PERCENT_N, NULL);
            case 'e':
            case 'E':
            case 'f':
            case 'g':
            case 'G':
                if (modifiers && float_expected == -1)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                if (!modifiers)
                    modifiers = sizeof(double);
                goto process_arg;
            case 'd':
            case 'i':
            case 'o':
            case 'u':
            case 'x':
            case 'X':
                if (float_expected == 1)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                if (!modifiers)
                    modifiers = sizeof(int);

process_arg:
                if (argnum)
                {
                    if (argnum > nextarg)
                        maxarg = argnum;
                    if (argument_sizes[argnum - 1] && argument_sizes[argnum - 1] != modifiers)
                        XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                    argument_sizes[argnum - 1] = modifiers;
                    nextarg = argnum; /* This might confuse, but it's right... */
                    break;
                }
                if (nextarg > SAFESTR_FMTARGS)
                    XXL_THROW_ERROR(SAFESTR_ERROR_TOO_MANY_FORMAT_ARGS, NULL);
                if (argument_sizes[nextarg] && argument_sizes[nextarg] != modifiers)
                    XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                argument_sizes[nextarg++] = modifiers;
                break;
            case '%':
                break;
            default:
                XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
                break;
        }
    }

    if (nextarg > maxarg)
        maxarg = nextarg;
    for (i = 0;  i < maxarg;  i++)
    {
        if (!argument_sizes[i])
            XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_STRING, NULL);
        if (argument_sizes[i] == -1)
        {
            popc = va_arg(ap, safestr_t);
            str = (isafestr_t)((char *)popc - sizeof(small_isafestr_t));
            if (str->hdr.cookie != safestr_cookie)
                XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_ARG, NULL);
            if (!(str->hdr.flags & SAFESTR_TRUSTED))
                trusted = 0;
            continue;
        }
        if (argument_sizes[i] % sizeof(void *))
            XXL_THROW_ERROR(SAFESTR_ERROR_INVALID_FORMAT_ARG, NULL);
        while (argument_sizes[i])
        {
            popi = (int)va_arg(ap, void *);
            argument_sizes[i] -= sizeof(void *);
        }
    }
    return trusted;
}

safestr_t
safestr_alloc(u_int32_t length, u_int32_t flags)
{
    u_int32_t   size;
    isafestr_t  str;

    /* Is the length requested too big? */
    if (length + 1 > (1 << (sizeof(u_int32_t) * 4)) - sizeof(small_isafestr_t))
        XXL_THROW_ERROR(SAFESTR_ERROR_STRING_TOO_LONG, NULL);

    size   = SAFESTR_ROUND(sizeof(small_isafestr_t) + length + 1);
    flags &= (SAFESTR_TEMPORARY | SAFESTR_IMMUTABLE | SAFESTR_TRUSTED);

    str = (isafestr_t)safestr_malloc(size, XXL_ASSET_PERMANENT);
    str->hdr.size   = size - sizeof(small_isafestr_t) - 1;
    str->hdr.length = 0;
    str->hdr.flags  = flags;
    str->hdr.cookie = get_cookie();
    str->str[0]     = 0;

    INC_REF_COUNT(str);
    return (safestr_t)str->str;
}

char
safestr_charat(safestr_t s, u_int32_t i)
{
    char        result;
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);
        if (i >= str->hdr.length)
            XXL_THROW_ERROR(SAFESTR_ERROR_INDEX_OUT_OF_RANGE, NULL);
        result = str->str[i];
    }
    XXL_ASSET_BLOCK_END;

    return result;
}

safestr_t
safestr_clone(safestr_t s, u_int32_t flags)
{
    u_int32_t   size;
    isafestr_t  clone, str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);

        /* Taint flag inheritance is independent of flags.  Can't pass taint by
         * calling _rw, because we have no need to check immutable.
         */
        flags &= ~SAFESTR_TRUSTED;
        flags |= (str->hdr.flags & SAFESTR_TRUSTED);

        size = SAFESTR_ROUND(sizeof(small_isafestr_t) + str->hdr.length + 1);
        clone = (isafestr_t)safestr_malloc(size, XXL_ASSET_PERMANENT);
        clone->hdr.size   = size - sizeof(small_isafestr_t) - 1;
        clone->hdr.length = str->hdr.length;
        clone->hdr.flags  = flags;
        clone->hdr.cookie = get_cookie();
        memcpy(clone->str, str->str, str->hdr.length + 1);
        INC_REF_COUNT(clone);
    }
    XXL_ASSET_BLOCK_END;

    return (safestr_t)clone->str;
}

int
safestr_compare(safestr_t s1, safestr_t s2, u_int32_t flags, ...)
{
    va_list     ap;
    u_int32_t   length = -1;

    if (flags & SAFESTR_COMPARE_LIMIT)
    {
        va_start(ap, flags);
        length = va_arg(ap, u_int32_t);
        va_end(ap);
    }
    return basecmp(s1, s2, flags, length);
}

void
safestr_concatenate(safestr_t *dst, safestr_t s, u_int32_t flags, ...)
{
    va_list     ap;
    u_int32_t   length, tlength;
    isafestr_t  str_dst, str_src;

    XXL_ASSET_BLOCK_BEGIN
    {
        str_src = get_isafestr_ro(s);
        str_dst = get_isafestr_rw(*dst, (str_src->hdr.flags & SAFESTR_TRUSTED));

        length = str_src->hdr.length;
        if (flags & SAFESTR_COPY_LIMIT)
        {
            va_start(ap, flags);
            tlength = va_arg(ap, u_int32_t);
            va_end(ap);
            length = SAFESTR_MIN(length, tlength);
        }

        str_dst = resize_safestr(str_dst, str_dst->hdr.length + length);
        length = SAFESTR_MIN(length, str_dst->hdr.size);
        memcpy(str_dst->str + str_dst->hdr.length, str_src->str, length);
        str_dst->hdr.length += length;
        *(str_dst->str + str_dst->hdr.length) = 0;
        *dst = (safestr_t)str_dst->str;
    }
    XXL_ASSET_BLOCK_END;
}

safestr_t *
safestr_convertarray(char **arr, u_int32_t flags)
{
    int         c, i;
    safestr_t   *res;
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        for (c = 0;  arr[c];  c++);
        res = (safestr_t *)safestr_malloc(sizeof(safestr_t) * (c + 1), XXL_ASSET_PROMOTE);

        for (i = 0;  arr[i];  i++)
        {
            res[i] = safestr_create(arr[i], flags);
            str = (isafestr_t)((char *)res[i] - sizeof(small_isafestr_t));
            XXL_ASSET_SAVE(str, free_isafestr_asset, NULL, XXL_ASSET_PROMOTE);
        }
        res[i] = NULL;
    }
    XXL_ASSET_BLOCK_END;

    return res;
}

safestr_t
safestr_create(char *s, u_int32_t flags)
{
    safestr_t   str;
    u_int32_t   length;
    isafestr_t  istr;

    length = strlen(s);
    str = safestr_alloc(length, flags);
    memcpy((char *)str, s, length + 1);
    istr = (isafestr_t)((char *)str - sizeof(small_isafestr_t));
    istr->hdr.length = length;
    return str;
}

void
safestr_delete(safestr_t *s, u_int32_t pos, u_int32_t count)
{
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_rw(*s, -1);

        if (pos < str->hdr.length)
        {
            if (pos + count >= str->hdr.length)
            {
                str->hdr.length = pos;
                *(str->str + str->hdr.length) = 0;
            }
            else
            {
                str->hdr.length -= count;
                memmove(str->str + pos, str->str + pos + count, 
                        str->hdr.length - pos + 1);
            }
        }
        *s = (safestr_t)str->str;
    }
    XXL_ASSET_BLOCK_END;
}

void
safestr_duplicate(safestr_t *dst, safestr_t src, u_int32_t flags, ...)
{
    va_list     ap;
    u_int32_t   length;
    isafestr_t  str_dst, str_src;

    XXL_ASSET_BLOCK_BEGIN
    {
        str_src = get_isafestr_ro(src);
        str_dst = get_isafestr_rw(*dst, (str_src->hdr.flags & SAFESTR_TRUSTED));

        if (!(flags & SAFESTR_COPY_LIMIT))
            length = str_src->hdr.length;
        else
        {
            va_start(ap, flags);
            length = va_arg(ap, u_int32_t);
            va_end(ap);
            length = SAFESTR_MIN(length, str_src->hdr.length);
        }

        str_dst = resize_safestr(str_dst, length);
        str_dst->hdr.length = length;
        memcpy(str_dst->str, str_src->str, length);
        *(str_dst->str + length) = 0;
        *dst = (safestr_t)str_dst->str;
    }
    XXL_ASSET_BLOCK_END;
}

int
safestr_equal(safestr_t s1, safestr_t s2, u_int32_t flags, ...)
{
    va_list     ap;
    u_int32_t   length = -1;

    if (flags & SAFESTR_COMPARE_LIMIT)
    {
        va_start(ap, flags);
        length = va_arg(ap, u_int32_t);
        va_end(ap);
    }

    return !basecmp(s1, s2, flags, length);
}

int
safestr_fprintf(FILE *stream, safestr_t fmt, ...)
{
    int         ret;
    va_list     ap;
    isafestr_t  format;

    XXL_ASSET_BLOCK_BEGIN
    {
        format = get_isafestr_ro(fmt);

        va_start(ap, fmt);
        check_format_string(format, ap);
        ret = vfprintf(stream, format->str, ap);
        va_end(ap);
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

void
safestr_free(safestr_t s)
{
    isafestr_t str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);
        DEC_REF_COUNT(str);
    }
    XXL_ASSET_BLOCK_END;
}

void
safestr_insert(safestr_t *dst, u_int32_t pos, safestr_t s)
{
    isafestr_t  str_dst, str_src;

    XXL_ASSET_BLOCK_BEGIN
    {
        str_src = get_isafestr_ro(s);
        str_dst = get_isafestr_rw(*dst, -1);

        str_dst->hdr.flags &= (str_src->hdr.flags & SAFESTR_TRUSTED);
        if (pos > str_dst->hdr.length)
            XXL_THROW_ERROR(SAFESTR_ERROR_INDEX_OUT_OF_RANGE, NULL);
        str_dst = resize_safestr(str_dst, str_dst->hdr.length + str_src->hdr.length);
        memmove(str_dst->str + pos + str_src->hdr.length, str_dst->str + pos, str_src->hdr.length);
        memcpy(str_dst->str + pos, str_src->str, str_src->hdr.length);
        str_dst->hdr.length += str_src->hdr.length;
        *(str_dst->str + str_dst->hdr.length) = '\0';
        *dst = (safestr_t)str_dst->str;
    }
    XXL_ASSET_BLOCK_END;
}

safestr_t
safestr_join(safestr_t *s, safestr_t joiner)
{
    char        *outp;
    safestr_t   ret;
    u_int32_t   i, count, len, trust = SAFESTR_TRUSTED;
    isafestr_t  ijoiner, istr, *strs;

    for (count = 0;  s[count];  count++);
    if (!count)
        return safestr_alloc(0, SAFESTR_TRUSTED);

    XXL_ASSET_BLOCK_BEGIN
    {
        strs = (isafestr_t *)safestr_malloc(sizeof(isafestr_t) * count, XXL_ASSET_TEMPORARY);

        ijoiner = get_isafestr_ro(joiner);
        trust  &= ijoiner->hdr.flags;

        for (i = len = 0;  i < count;  i++)
        {
            strs[i] = get_isafestr_ro(s[i]);
            len    += strs[i]->hdr.length;
            trust  &= strs[i]->hdr.flags;
        }
        len += (ijoiner->hdr.length * (count - 1));
        ret  = safestr_alloc(len, trust);
        istr = (isafestr_t)((char *)ret - sizeof(small_isafestr_t));
        istr->hdr.length = len;
        istr->str[len] = '\0';
        memcpy(istr->str, strs[0]->str, strs[0]->hdr.length);
        for (outp = istr->str + strs[0]->hdr.length, i = 1;  i < count;  i++)
        {
            memcpy(outp, ijoiner->str, ijoiner->hdr.length);
            memcpy(outp + ijoiner->hdr.length, strs[i]->str, strs[i]->hdr.length);
            outp += (ijoiner->hdr.length + strs[i]->hdr.length);
        }
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

u_int32_t
safestr_length(safestr_t s)
{
    u_int32_t   length;
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);
        length = str->hdr.length;
    }
    XXL_ASSET_BLOCK_END;

    return length;
}

void 
safestr_memzero(volatile void *str, u_int32_t len)
{
    volatile char   *buf;

    for (buf = (volatile char *)str;  len;  buf[--len] = 0);
}

int
safestr_printf(safestr_t fmt, ...)
{
    int         ret;
    va_list     ap;
    isafestr_t  format;

    XXL_ASSET_BLOCK_BEGIN
    {
        format = get_isafestr_ro(fmt);
        va_start(ap, fmt);
        check_format_string(format, ap);
        ret = vprintf(format->str, ap);
        va_end(ap);
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

u_int32_t
safestr_reference(safestr_t s)
{
    u_int32_t   result;
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);
        result = REF_COUNT(str);
        INC_REF_COUNT(str);
    }
    XXL_ASSET_BLOCK_END;

    return result;
}

safestr_t
safestr_replace(safestr_t str, safestr_t oldp, safestr_t newp)
{
    char        *c, *outp;
    safestr_t   result;
    u_int32_t   new_len, trusted;
    isafestr_t  new_pat, new_str, old_pat, src_str;

    XXL_ASSET_BLOCK_BEGIN
    {
        old_pat = get_isafestr_ro(oldp);
        src_str = get_isafestr_ro(str);
        if (src_str->hdr.length < old_pat->hdr.length)
            result = safestr_clone(str, (src_str->hdr.flags & SAFESTR_TRUSTED));
        else
        {
            new_pat = get_isafestr_ro(newp);
            new_len = 0;
            for (c = src_str->str;  *(c + old_pat->hdr.length - 1);  c++)
            {
                if (memcmp(c, old_pat->str, old_pat->hdr.length))
                {
                    new_len++;
                    continue;
                }
                new_len += new_pat->hdr.length;
                c += (old_pat->hdr.length) - 1;
            }
            if (*c)
                new_len += (old_pat->hdr.length - 1);

            trusted  = src_str->hdr.flags & SAFESTR_TRUSTED;
            trusted &= new_pat->hdr.flags & SAFESTR_TRUSTED;
            result   = safestr_alloc(new_len, trusted);
            new_str  = (isafestr_t)((char *)result - sizeof(small_isafestr_t));
            outp     = new_str->str;
            XXL_ASSET_SAVE(new_str, free_isafestr_asset, NULL, XXL_ASSET_PROMOTE);

            for (c = src_str->str;  *(c + old_pat->hdr.length - 1);  c++)
            {
                if (memcmp(c, old_pat->str, old_pat->hdr.length))
                {
                    *outp++ = *c;
                    continue;
                }
                memcpy(outp, new_pat->str, new_pat->hdr.length);
                outp += new_pat->hdr.length;
                c += old_pat->hdr.length - 1;
            }
            if (*c)
            {
                memcpy(outp, c, old_pat->hdr.length - 1);
                outp += (old_pat->hdr.length - 1);
            }
            *outp = '\0';
            new_str->hdr.length =  new_len;
        }
    }
    XXL_ASSET_BLOCK_END;

    return result;
}

u_int32_t
safestr_search(safestr_t s, safestr_t sub, u_int32_t flags, ...)
{
    int        from = 0, n = 1, result;
    char       *ptr, *in, *what;
    va_list    ap;
    u_int32_t  fromchar = 0, len;
    isafestr_t str, substr;
    
    XXL_ASSET_BLOCK_BEGIN
    {
        str    = get_isafestr_ro(s);
        substr = get_isafestr_ro(sub);

        va_start(ap, flags);
        if (flags & SAFESTR_FIND_FROMCHAR)
            fromchar = va_arg(ap, int);
        if (flags & SAFESTR_FIND_FROMNTH)
            from = va_arg(ap, int);
        if (flags & SAFESTR_FIND_NTH)
            n = va_arg(ap, int);
        va_end(ap);

        n += from;
        if (n < 1 || fromchar >= str->hdr.length || 
            substr->hdr.length > str->hdr.length)
        {
            result = SAFESTR_ERROR_NOTFOUND;
        }
        else
        {
            if (flags & SAFESTR_FIND_NOMATCHCASE)
            {
                in = safestr_strdup(str->str);
                XXL_ASSET_SAVE(in, free_void_asset, NULL, XXL_ASSET_TEMPORARY);
#ifndef WIN32
                for (ptr = in;  *ptr;  ptr++)
                    *ptr = tolower(*ptr);
#else
                CharLowerBuffA(in, lstrlenA(in));
#endif
  
                what = safestr_strdup(substr->str);
                XXL_ASSET_SAVE(what, free_void_asset, NULL, XXL_ASSET_TEMPORARY);
#ifndef WIN32
                for (ptr = what;  *ptr;  ptr++)
                    *ptr = tolower(*ptr);
#else
                CharLowerBuffA(what, lstrlenA(what));
#endif
            }
            else
            {
                in   = str->str;
                what = substr->str;
            }

            if (flags & SAFESTR_FIND_REVERSE)
            {
                len = strlen(what);
                for (ptr = in + strlen(in) - fromchar - len;  ptr >= in;  ptr--)
                    if (!strncmp(ptr, what, len) && !--n)
                        break;
                if (ptr < in)
                    ptr = NULL;
            }
            else
            {
                ptr = strstr(in + fromchar, what);
                while (--n && ptr)
                    ptr = strstr(ptr + 1, what);
            }

            result = (!ptr ? SAFESTR_ERROR_NOTFOUND : (ptr - in));
        }
    }
    XXL_ASSET_BLOCK_END;

    return result;
}

void
safestr_setcharat(safestr_t s, u_int32_t i, char c, int trust)
{
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_rw(s, trust);
        if (i >= str->hdr.length)
            XXL_THROW_ERROR(SAFESTR_ERROR_INDEX_OUT_OF_RANGE, NULL);
        str->str[i] = c;
    }
    XXL_ASSET_BLOCK_END;
}

void
safestr_setmemfns(safestr_malloc_t malloc_fn, safestr_realloc_t realloc_fn,
                  safestr_free_t free_fn)
{
    safestr_malloc_fn  = (malloc_fn  ? malloc_fn  : malloc);
    safestr_realloc_fn = (realloc_fn ? realloc_fn : realloc);
    safestr_free_fn    = (free_fn    ? free_fn    : free);
}

safestr_t
safestr_slice(safestr_t s, u_int32_t start, u_int32_t end)
{
    u_int32_t   length = 0, size;
    isafestr_t  new_str, str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_ro(s);
        if (start < str->hdr.length)
        {
            length  = end - start;
            length  = SAFESTR_MIN(str->hdr.length - start, length);
        }

        size = SAFESTR_ROUND(sizeof(small_isafestr_t) + length + 1);
        new_str = (isafestr_t)safestr_malloc(size, XXL_ASSET_PERMANENT);
        new_str->hdr.size    = size - sizeof(small_isafestr_t) - 1;
        new_str->hdr.length  = length;
        new_str->hdr.flags   = (str->hdr.flags & SAFESTR_TRUSTED);
        new_str->hdr.cookie  = get_cookie();
        new_str->str[length] = 0;
        memcpy(new_str->str, str->str + start, length);
        INC_REF_COUNT(new_str);
    }
    XXL_ASSET_BLOCK_END;

    return (safestr_t)new_str->str;
}

safestr_t *
safestr_split(safestr_t s, safestr_t sub)
{
    char        *c, *cursor;
    safestr_t   *ret;
    u_int32_t   count;
    isafestr_t  istr, str, substr;

    XXL_ASSET_BLOCK_BEGIN
    {
        str    = get_isafestr_ro(s);
        substr = get_isafestr_ro(sub);

        if (!substr->hdr.length)
            ret = charlist(str);
        else if (substr->hdr.length > str->hdr.length)
        {
            ret = (safestr_t *)safestr_malloc(sizeof(safestr_t) * 2, XXL_ASSET_PROMOTE);
            ret[0] = safestr_clone(s, (str->hdr.flags & SAFESTR_TRUSTED));
            ret[1] = NULL;
        }
        else
        {
            /* Figure out how many array elements are needed */
            count = 1;
            for (c = str->str;  *(c + substr->hdr.length - 1);  c++)
                if (!memcmp(c, substr->str, substr->hdr.length))
                {
                    count++;
                    c += (substr->hdr.length - 1);
                }

            ret = (safestr_t *)safestr_malloc(sizeof(safestr_t) * (count + 1), XXL_ASSET_PROMOTE);

            count = 0;
            for (c = cursor = str->str;  *(c + substr->hdr.length - 1);  c++)
            {
                if (!memcmp(c, substr->str, substr->hdr.length))
                {
                    ret[count] = safestr_alloc(c - cursor, (str->hdr.flags & SAFESTR_TRUSTED));
                    istr = (isafestr_t)((char *)ret[count] - sizeof(small_isafestr_t));
                    XXL_ASSET_SAVE(istr, free_isafestr_asset, NULL, XXL_ASSET_PROMOTE);
                    istr->hdr.length = c - cursor;
                    memcpy(istr->str, cursor, istr->hdr.length);
                    istr->str[istr->hdr.length] = '\0';
                    cursor = c + substr->hdr.length;
                    c += (substr->hdr.length - 1);
                    count++;
                }
            }
            if (*c)
                c += (substr->hdr.length - 1);
            ret[count] = safestr_alloc(c - cursor, (str->hdr.flags & SAFESTR_TRUSTED));
            istr = (isafestr_t)((char *)ret[count] - sizeof(small_isafestr_t));
            XXL_ASSET_SAVE(istr, free_isafestr_asset, NULL, XXL_ASSET_PROMOTE);
            istr->hdr.length = c - cursor;
            memcpy(istr->str, cursor, istr->hdr.length);
            istr->str[istr->hdr.length] = '\0';
            ret[++count] = NULL;
        }
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

#ifdef HAVE_VSNPRINTF
u_int32_t
safestr_sprintf(safestr_t *s, safestr_t fmt, ...)
{
    int         trusted;
    va_list     ap;
    u_int32_t   need;
    isafestr_t  format, str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str    = get_isafestr_rw(*s, 0);
        format = get_isafestr_ro(fmt);

        str->hdr.flags &= ~SAFESTR_TRUSTED;
        va_start(ap, fmt);
        trusted = check_format_string(format, ap);
        need = vsnprintf(str->str, str->hdr.size + 1, format->str, ap);
        va_end(ap);

        if (need > str->hdr.size)
        {
            str = resize_safestr(str, need);
            va_start(ap, fmt);
            need = vsnprintf(str->str, str->hdr.size + 1, format->str, ap);
            va_end(ap);
        }

        str->hdr.length = need;
        if (trusted && (format->hdr.flags & SAFESTR_TRUSTED))
            str->hdr.flags |= SAFESTR_TRUSTED;
        *s = (safestr_t)str->str;
    }
    XXL_ASSET_BLOCK_END;
  
    return need;
}
#endif

char *
safestr_strdup(char *s)
{
    char        *new_s;
    u_int32_t   len;

    len = strlen(s) + 1;
    new_s = (char *)safestr_malloc(len, XXL_ASSET_PERMANENT);
    memcpy(new_s, s, len);
    return new_s;
}

void
safestr_truncate(safestr_t *s, u_int32_t length)
{
    isafestr_t  str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str = get_isafestr_rw(*s, -1);
        if (length > str->hdr.length)
        {
            str = resize_safestr(str, length);
            /* pad the new space with spaces */
            memset(str->str + str->hdr.length, ' ', length - str->hdr.length);
        }
        str->hdr.length = length;
        *(str->str + str->hdr.length) = 0;
        *s = (safestr_t)str->str;
    }
    XXL_ASSET_BLOCK_END;
}

int
safestr_vfprintf(FILE *stream, safestr_t fmt, va_list ap)
{
    int         ret;
    isafestr_t  format;

    XXL_ASSET_BLOCK_BEGIN
    {
        format = get_isafestr_ro(fmt);
        check_format_string(format, ap);
        ret = vfprintf(stream, format->str, ap);
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

int
safestr_vprintf(safestr_t fmt, va_list ap)
{
    int         ret;
    isafestr_t  format;

    XXL_ASSET_BLOCK_BEGIN
    {
        format = get_isafestr_ro(fmt);
        check_format_string(format, ap);
        ret = vprintf(format->str, ap);
    }
    XXL_ASSET_BLOCK_END;

    return ret;
}

#ifdef HAVE_VSNPRINTF
u_int32_t
safestr_vsprintf(safestr_t *s, safestr_t fmt, va_list ap)
{
    int         trusted;
    u_int32_t   need;
    isafestr_t  format, str;

    XXL_ASSET_BLOCK_BEGIN
    {
        str    = get_isafestr_rw(*s, 0);
        format = get_isafestr_ro(fmt);

        trusted = check_format_string(format, ap);
        str->hdr.flags &= ~SAFESTR_TRUSTED;
        if (trusted)
            str->hdr.flags |= SAFESTR_TRUSTED;
        need = vsnprintf(str->str, str->hdr.size + 1, format->str, ap);

        if (need > str->hdr.size)
        {
            str = resize_safestr(str, need);
            need = vsnprintf(str->str, str->hdr.size + 1, format->str, ap);
        }

        str->hdr.length = need;
        *s = (safestr_t)str->str;
    }
    XXL_ASSET_BLOCK_END;

    return need;
}
#endif

/* ------------------------------------------------------------------------- */

void
safestr_trust(safestr_t s)
{
    XXL_ASSET_BLOCK_BEGIN
    {
        get_isafestr_ro(s)->hdr.flags |= SAFESTR_TRUSTED;
    }
    XXL_ASSET_BLOCK_END;
}

void
safestr_untrust(safestr_t s)
{
    XXL_ASSET_BLOCK_BEGIN
    {
        get_isafestr_ro(s)->hdr.flags &= ~SAFESTR_TRUSTED;
    }
    XXL_ASSET_BLOCK_END;
}

int
safestr_istrusted(safestr_t s)
{
    u_int32_t   trusted;

    XXL_ASSET_BLOCK_BEGIN
    {
        trusted = get_isafestr_ro(s)->hdr.flags & SAFESTR_TRUSTED;
    }
    XXL_ASSET_BLOCK_END;

    return !!trusted;
}

void 
safestr_makereadonly(safestr_t s)
{
    XXL_ASSET_BLOCK_BEGIN
    {
        get_isafestr_ro(s)->hdr.flags |= SAFESTR_IMMUTABLE;
    }
    XXL_ASSET_BLOCK_END;
}

void
safestr_makewritable(safestr_t s)
{
    XXL_ASSET_BLOCK_BEGIN
    {
        get_isafestr_ro(s)->hdr.flags &= ~SAFESTR_IMMUTABLE;
    }
    XXL_ASSET_BLOCK_END;
}

int
safestr_isreadonly(safestr_t s)
{
    u_int32_t   immutable;

    XXL_ASSET_BLOCK_BEGIN
    {
        immutable = get_isafestr_ro(s)->hdr.flags & SAFESTR_IMMUTABLE;
    }
    XXL_ASSET_BLOCK_END;

    return !!immutable;
}
